<?php

	require_once("_parameters.php");
	require_once __DIR__ . '/lib/mathphp/LinearAlgebra/Vector.php';
	define("ORIGINAL_STRAIN", 1);
	define("PATH", "files");
	define("DEFAULT_WINDOW_SIZE_PERCENTAGE", 1);
	
	use MathPHP\LinearAlgebra\Vector;
	
	class Genome
	{
		private $_accession;
		private $_length;
		private $_inflectionPoints;
		private $_ATvalues;
		private $_GCvalues;
		private $_CGCvalues;
		private $_CATvalues;
		private $_gCount;
		private $_cCount;
		private $_aCount;
		private $_tCount;
		private $_fftnumericSequence;
		private $_inflectionPointsvalues;
		private $_windowValues;
		private $_numberOfWindows;

		public function __construct()
		{
			$this->_gCount = 0;
			$this->_cCount = 0;
			$this->_aCount = 0;
			$this->_tCount = 0;
			$this->_fftnumericSequence = [];
			$this->_calculateFileDetails();
		}
		
		public function GetATvaluesForGraph()
		{
			return $this->_ATvalues;
		}
		
		public function GetGCvaluesForGraph()
		{
			return $this->_GCvalues;
		}

		public function GetCGCvaluesForGraph()
		{
			return $this->_CGCvalues;
		}
		
		public function GetCATvaluesForGraph()
		{
			return $this->_CATvalues;
		}
		
		public function GetInflectionPointsValuesForGraph()
		{
			return $this->_inflectionPointsvalues;
		}
		
		public function GetWindowValues()
		{
			return $this->_windowValues;
		}
		
		public function GetNumberOfWindows()
		{
			return $this->_numberOfWindows;
		}
		
		public function CheckKnockoutCriteria()
		{
			return $this->_checkKOCriteria();
		}
		
		// Global check - GC content in %; entropy
		//	return values: true - the sequence could contain oriC; false - the sequence most likely doesn't contain the oriC
		public function GlobalCheck()
		{
			$gcContent = $this->_getGCContent();
			$entropy = $this->_calculateEntropy();
			$skews = $this->_calculateGenomeSkews();
			$variance = $this->_nucleotideVariance();
			
			$checks = [
				$gcContent >= GC_THRESHOLD,
				$entropy <= ENTROPY_THRESHOLD,
				(abs($skews["gcSkew"]) >= SKEW_THRESHOLD),
				$variance >= VARIANCE_THRESHOLD
			];

			$positiveChecks = array_sum($checks);
			$isValid = $this->_checkKOCriteria($gcContent);
			if ($isValid)
			{
				$isValid = $positiveChecks >= TOTAL_POSITIVE_CHECKS_THRESHOLD;
			}

			return ["isValid" => $isValid, "gcContent" => $gcContent, "entropy" => $entropy, "gcSkew" => abs($skews["gcSkew"]), "variance" => $variance];
		}
		
		public function _checkKOCriteria()
		{
			$isValid = false;
			$gcContent = $this->_getGCContent();
			$isGCValid = $gcContent > GC_KNOCKOUT;
			$isLengthValid = $this->_length > LENGTH_KNOCKOUT;
			if ($isGCValid && $isLengthValid)
			{
				$isValid = true;
			}
			return ['isValid' => $isValid, 'gcContent' => $gcContent, 'length' => $this->_length];
		}
		
		public function CalculateSkews($windowSizePercantage = DEFAULT_WINDOW_SIZE_PERCENTAGE)
		{
			$windowSize = floor($windowSizePercantage * $this->_length / 100);
			$result = $this->_calculateSkews($windowSize);

			//$numberOfwindows = count($result["calculation"]);
			//$this->_numberOfWindows = $numberOfwindows;
			$summary = $result["summary"];
			
			$result["windowSizePercentage"] = $windowSizePercantage;
			$result["windowSize"] = $windowSize;
			
			UNSET($summary);
			return $result;
		}

		private function _getGCContent()
		{
			$gcContent = 0;
			if (ISSET($this->_length) && $this->_length > 0)
			{
				$gcContent = ((($this->_gCount + $this->_cCount) / $this->_length) * 100);
			}
			return $gcContent;
		}
		
		private function _calculateEntropy()
		{
			$entropy = 0;
			
			if (ISSET($this->_length) && $this->_length > 0)
			{
				$frequencies = [
					'A' => $this->_aCount / $this->_length,
					'T' => $this->_tCount / $this->_length,
					'G' => $this->_gCount / $this->_length,
					'C' => $this->_cCount / $this->_length
				];

				foreach ($frequencies as $freq) 
				{
					if ($freq > 0) 
					{
						$entropy += $freq * log($freq, 2); // Shannonova formula
					}
				}
			}

			return -$entropy; // Entropy is a negative value, always returning a positive value
		}
		
		private function _nucleotideVariance()
		{
			
			$counts = [$this->_gCount, $this->_aCount, $this->_tCount, $this->_cCount];

			$mean = array_sum($counts) / count($counts);
			
			$variance = 0;
			foreach ($counts as $count) 
			{
				$variance += pow($count - $mean, 2);
			}

			return $variance / count($counts); 
		}
		
		private function _calculateGenomeSkews()
		{
			$gcSkew = ($this->_gCount - $this->_cCount) / ($this->_gCount + $this->_cCount + 1e-10);
			$atSkew = ($this->_aCount - $this->_tCount) / ($this->_aCount + $this->_tCount + 1e-10);
			
			return ['gcSkew' => $gcSkew, 'atSkew' => $atSkew];
		}
		
		private function _calculateSkews($windowSize)
		{
			$filename = $_FILES['sequenceInput']['tmp_name'];
			if(!file_exists($filename))
			{
				echo "[1001] Unable to load file!";
				return null;
			}
			
			$myfile = fopen($filename, "r");
			if (!$myfile) 
			{
				echo "[1002] Unable to open file!";
				return null;
			}
			
			$firstRow = fgets($myfile);
		
			$i = 1;
			$window = 1;
			$this->_windowValues = "[0,";
			$G = 0;
			$C = 0;
			$A = 0;
			$T = 0;
			$result[0]["cGCskew"] = 0;
			$result[0]["cATskew"] = 0;
			$this->_ATvalues = "[";
			$this->_GCvalues = "[";
			$this->_CGCvalues = "[";
			$this->_CATvalues = "[";
			
			while(!feof($myfile)) 
			{
				// if the whole window is loaded, calculate the data
				if ($i % $windowSize == 0)
				{
					$this->_windowValues .= sprintf("%d, ", $i);
					$GC = ($G - $C)/(float)($G + $C);
					$AT = ($A - $T)/(float)($A + $T);
					$result[$window]["G"] = $G;
					$result[$window]["C"] = $C;
					$result[$window]["A"] = $A;
					$result[$window]["T"] = $T;
					//every based value is Y-axes value
					/*
						calculating q_i = (y_i+1 - y_i)/(x_i+1 - x_i)
						based on q_i, p_i is calculated
						p_i = (q_i+1 - q_i) / (x_i+1 - x_i)
						inflection point is p_i * p_i+1 <= 0 for the point x_i+2
					*/
					$result[$window]["GCskew"] = $GC;
					$result[$window]["ATskew"] = $AT;
					$result[$window]["cGCskew"] = $result[$window-1]["cGCskew"] + $GC;
					$result[$window]["cATskew"] = $result[$window-1]["cATskew"] + $AT;
					
					$this->_ATvalues .= sprintf("%f, ", $result[$window]["ATskew"]);
					$this->_GCvalues .= sprintf("%f, ", $result[$window]["GCskew"]);
					$this->_CGCvalues .= sprintf("%f, ", $result[$window]["cGCskew"]);
					$this->_CATvalues .= sprintf("%f, ", $result[$window]["cATskew"]);
					
					if ($window == 1)
					{
						$summary["minAT"]["value"] = $AT;
						$summary["minAT"]["positionLeft"] = 0;
						$summary["minAT"]["positionRight"] = 1*$windowSize;
						
						$summary["maxAT"]["value"] = $AT;
						$summary["maxAT"]["positionLeft"] = 0;
						$summary["maxAT"]["positionRight"] = 1*$windowSize;
						
						$summary["minGC"]["value"] = $GC;
						$summary["minGC"]["positionLeft"] = 0;
						$summary["minGC"]["positionRight"] = 1*$windowSize;
						
						$summary["maxGC"]["value"] = $GC;
						$summary["maxGC"]["positionLeft"] = 0;
						$summary["maxGC"]["positionRight"] = 1*$windowSize;
						
						$summary["mincAT"]["value"] = $result[$window]["cATskew"];
						$summary["mincAT"]["positionLeft"] = 0;
						$summary["mincAT"]["positionRight"] = 1*$windowSize;
						
						$summary["maxcAT"]["value"] = $result[$window]["cATskew"];
						$summary["maxcAT"]["positionLeft"] = 0;
						$summary["maxcAT"]["positionRight"] = 1*$windowSize;
						
						$summary["mincGC"]["value"] = $result[$window]["cGCskew"];
						$summary["mincGC"]["positionLeft"] = 0;
						$summary["mincGC"]["positionRight"] = 1*$windowSize;
						
						$summary["maxcGC"]["value"] = $result[$window]["cGCskew"];
						$summary["maxcGC"]["positionLeft"] = 0;
						$summary["maxcGC"]["positionRight"] = 1*$windowSize;
					}
					else
					{
						if ($AT < $summary["minAT"]["value"])
						{
							$summary["minAT"]["value"] = $AT;
							$summary["minAT"]["positionLeft"] = ($window-1)*$windowSize;
							$summary["minAT"]["positionRight"] = $window*$windowSize;
						}
						
						if ($AT > $summary["maxAT"]["value"])
						{
							$summary["maxAT"]["value"] = $AT;
							$summary["maxAT"]["positionLeft"] = ($window-1)*$windowSize;
							$summary["maxAT"]["positionRight"] = $window*$windowSize;
						}
						
						if ($GC < $summary["minGC"]["value"])
						{					
							$summary["minGC"]["value"] = $GC;
							$summary["minGC"]["positionLeft"] = ($window-1)*$windowSize;
							$summary["minGC"]["positionRight"] = $window*$windowSize;
						}
						
						if ($GC > $summary["maxGC"]["value"])
						{
							$summary["maxGC"]["value"] = $GC;
							$summary["maxGC"]["positionLeft"] = ($window-1)*$windowSize;
							$summary["maxGC"]["positionRight"] = $window*$windowSize;
						}
						
						if ($result[$window]["cATskew"] < $summary["mincAT"]["value"])
						{
							$summary["mincAT"]["value"] = $result[$window]["cATskew"];
							$summary["mincAT"]["positionLeft"] = ($window-1)*$windowSize;
							$summary["mincAT"]["positionRight"] = $window*$windowSize;
						}
						
						if ($result[$window]["cATskew"] > $summary["maxcAT"]["value"])
						{
							$summary["maxcAT"]["value"] = $result[$window]["cATskew"];
							$summary["maxcAT"]["positionLeft"] = ($window-1)*$windowSize;
							$summary["maxcAT"]["positionRight"] = $window*$windowSize;
						}
						
						if ($result[$window]["cGCskew"] < $summary["mincGC"]["value"])
						{
							$summary["mincGC"]["value"] = $result[$window]["cGCskew"];
							$summary["mincGC"]["positionLeft"] = ($window-1)*$windowSize;
							$summary["mincGC"]["positionRight"] = $window*$windowSize;
						}
						
						if ($result[$window]["cGCskew"] > $summary["maxcGC"]["value"])
						{
							$summary["maxcGC"]["value"] = $result[$window]["cGCskew"];
							$summary["maxcGC"]["positionLeft"] = ($window-1)*$windowSize;
							$summary["maxcGC"]["positionRight"] = $window*$windowSize;
						}
					}
					
					UNSET($GC);
					UNSET($AT);
					$Gcount = 0;
					$Ccount = 0;
					$Acount = 0;
					$Tcount = 0;
					$window++;
				}
								
				do {
					$currentBP = fgetc($myfile);
					//$currentBP = strtoupper($currentBP);
				} while ($currentBP !== false && !in_array($currentBP, array('A', 'T', 'G', 'C')));
				
				switch ($currentBP)
				{
					case "A":
					case "a":
						$A++;
						$i++;
						break;
					case "T":
					case "t":
						$T++;
						$i++;
						break;
					case "G":
					case "g":
						$G++;
						$i++;
						break;
					case "C":
					case "c":
						$C++;
						$i++;
						break;
					default:
						break;
				}
			}
			
			if (!ISSET($result[$window]) && ($C > 0 || $G>0 || $A>0 || $T>0))
			{
				$this->_windowValues .= sprintf("%d,", $i);
				$GC = ($G - $C)/(float)($G + $C);
				$AT = ($G - $C)/(float)($G + $C);
				$result[$window]["G"] = $G;
				$result[$window]["C"] = $C;
				$result[$window]["A"] = $A;
				$result[$window]["T"] = $T;
				$result[$window]["GCskew"] = $GC;
				$result[$window]["ATskew"] = $AT;
				$result[$window]["cGCskew"] = $result[$window-1]["cGCskew"] + $GC;
				$result[$window]["cATskew"] = $result[$window-1]["cATskew"] + $AT;
				
				$this->_ATvalues .= sprintf("%f", $result[$window]["GCskew"]);
				$this->_GCvalues .= sprintf("%f", $result[$window]["ATskew"]);
				$this->_CGCvalues .= sprintf("%f", $result[$window]["cGCskew"]);
				$this->_CATvalues .= sprintf("%f", $result[$window]["cATskew"]);
				
				$window++;
			}

			$this->_ATvalues .= "]";
			$this->_GCvalues .= "]";
			$this->_CGCvalues .= "]";
			$this->_CATvalues .= "]";
			$this->_windowValues = sprintf("%s]", rtrim($this->_windowValues, ','));
			
			//calculation of inflection points
			for($i= 1; $i < count($result) - 1; $i++)
			{
				$result[$i]["X"] = $i;
				// 
				$result[$i]["cGCskew_Q"] = ($result[$i+1]["cGCskew"] - $result[$i]["cGCskew"]) / (($windowSize * $i) - ($windowSize * ($i-1)));
			}

			for($i= 1; $i < count($result) - 2; $i++)
			{
				$result[$i]["cGCskew_P"] = ($result[$i+1]["cGCskew_Q"] - $result[$i]["cGCskew_Q"]) / (($windowSize * $i) - ($windowSize * ($i-1)));
			}

			$result[1]["cGCskew_PT"] = false;
			$result[2]["cGCskew_PT"] = false;
			$foundFirst = false;
			$this->_inflectionPointsvalues ="[";
		
			for($i= 3; $i < count($result); $i++)
			{
				$result[$i]["cGCskew_PT"] = false;
				
				if (ISSET($result[$i-2]["cGCskew_P"], $result[$i-1]["cGCskew_P"]) && ($result[$i-2]["cGCskew_P"] * $result[$i-1]["cGCskew_P"] < 0))
				{
					$result[$i]["cGCskew_PT"] = true;
					$inflectionPoint["numberOfWindow"] = $i;
					$inflectionPoint["positionLeft"] = $windowSize * ($i - 1);
					$inflectionPoint["positionRight"] = $windowSize * $i;
					$inflectionPoint["firstAfterMaxCGC"] = (($inflectionPoint["positionLeft"] > $summary["maxcGC"]["positionRight"]) && !$foundFirst) ? 1 : 0;
					$inflectionPoint["alignedWithMinGC"] = ($inflectionPoint["positionLeft"] == $summary["minGC"]["positionLeft"]) ? 1 : 0;
					if ($inflectionPoint["firstAfterMaxCGC"] == 1)
					{
						$foundFirst = true;
					}
					$inflectionPoints[] = $inflectionPoint;
					$this->_inflectionPointsvalues .= sprintf("%f,", $result[$i]["cGCskew"]);
				}
			}
			
			fclose($myfile);
			$this->_inflectionPoints = $inflectionPoints;
			$this->_inflectionPointsvalues = sprintf("%s]", rtrim($this->_inflectionPointsvalues, ','));
			$this->_numberOfWindows = $window-1;
			return array("calculation"=>$result, "summary"=>$summary, "inflectionPoints" => $inflectionPoints);
		}
		
		public function getSpecificInflectionPoint($mode = IP_FIRST_AFTER_MAXCGC)
		{
			$result = null; 

			foreach($this->_inflectionPoints as $inflectionPoint)
			{
				if ($mode == IP_FIRST_AFTER_MAXCGC && $inflectionPoint["firstAfterMaxCGC"] == 1)
				{
					$result = $inflectionPoint;
					break;
				}

				if ($mode == IP_ALIGNED_WITH_MINGC && $inflectionPoint["alignedWithMinGC"] == 1)
				{
					$result = $inflectionPoint;
					break;
				}
			}
			
			return $result;
		}
	
		private function _calculateFileDetails()
		{
			$filename = $_FILES['sequenceInput']['tmp_name'];
			$myfile = fopen($filename, "r") or die("[1003] Unable to open file!");
			$firstRow = fgets($myfile);

			$tmp = explode(">", $firstRow);
			$tmp = explode(" ", $tmp[1]);
			$tmp = explode(".", $tmp[0]);
			define("ACCESSION", trim($tmp[0]));

			$number = 0;
			while(!feof($myfile)) 
			{
				$currentBP = fgetc($myfile);
				switch ($currentBP)
				{
					case "A":
					case "a":
						$this->_aCount++;
						$this->_numericSequence[] = 1;
						$number++;
						break;
					case "T":
					case "t":
						$this->_tCount++;
						$this->_numericSequence[] = -1;
						$number++;
						break;
					case "G":
					case "g":
						$this->_gCount++;
						$this->_numericSequence[] = 1;
						$number++;
						break;
					case "C":
					case "c":
						$this->_cCount++;
						$this->_numericSequence[] = -1;
						$number++;
						break;
					default:
						break;
				}
			}

			define("LENGTH", $number);
			define("GENOME_SIZE", $number);

			$this->_accession = ACCESSION;
			$this->_length = LENGTH;
		}
	}


